#!/usr/local/bin/perl58 -w

###
# Title:        FASSI (Fingerprint and Assembly incorporation)
# Description:  Producing map-derived constraints for PCAP assembler
# Copyright:    Copyright (c) 2005, 2006
# Company:      Genome Sciences Centre, BC Cancer Agency
# Author:       Dmitry Varabei
# Version 1.0
###


use strict;
use Getopt::Long;
use vars qw(%OPT);


#general parameters. These are default values, there are using, if not defined in profile.

my $shared_length_to_count_overlap = 0.4;  #similarity to generate constraints between 2 neighbours
my $shared_length_to_count_overlap2 = 0.3;  #similarity to generate constraints between other neighbours
my $Dev = 0.3;   #add to constaints size;
my $default_BAC_length = 170000;
my ($default_constr_min_size,$default_constr_max_size)=(60000,300000);
my ($min_band_size, $max_band_size) = (0,30000);
my $multiplier = 1;
my ($tolerance_min_band,$tolerance_max_band) = (7,7);


# structures to store fingerprint information

my $fpc_clones = {};
my $fpc_contigs = ();
my $parent = ();
my $kids = ();
my $size_errors = ();
my $bands = {};
my $band_sizes = {};
my $num_same_position = 0;

# structures to store sequence information


my $sequence = {};
my $quality = {};
my ($size_errors_min,$size_errors_max) = (500,30000);
my $library_name;
my $reads_to_print;


# parameters list from profile
my $PARAM;



if (not defined($ARGV[0])){
  write;    
  exit;    
}

my $profile = $ARGV[0];

read_param($profile);

fassi_main();

exit;




#
# read parameters from config file
#

sub read_param {
   my $file = shift; 
   open (FilePar, $file) or die "can't open the file $file\n";
   while (<FilePar>) {
     if (/^\#/){
     } else { 
       if (/^(\S+)\s+=\s+(\S+)/){
        $PARAM->{$1} = $2;
       }
     }  
   }
   
#  set parameter values

   if (defined ($PARAM->{shared_length_to_count_overlap})){
     $shared_length_to_count_overlap = $PARAM->{shared_length_to_count_overlap};
   };
   if (defined ($PARAM->{shared_length_to_count_overlap2})){
     $shared_length_to_count_overlap2 = $PARAM->{shared_length_to_count_overlap2};
   };  
   if (defined ($PARAM->{Dev})){
     $Dev = $PARAM->{Dev};
   };  
   if (defined ($PARAM->{default_BAC_length})){
     $default_BAC_length = $PARAM->{default_BAC_length};
   };  
   if (defined ($PARAM->{default_constr_min_size})){
     $default_constr_min_size = $PARAM->{default_constr_min_size};
   };  
   if (defined ($PARAM->{default_constr_max_size})){
     $default_constr_max_size = $PARAM->{default_constr_max_size};
   };  
   if (defined ($PARAM->{min_band_size})){
     $min_band_size = $PARAM->{min_band_size};
   };
   if (defined ($PARAM->{max_band_size})){
     $max_band_size = $PARAM->{max_band_size};
   };
   if (defined ($PARAM->{multiplier})){
     $multiplier = $PARAM->{multiplier};
   };
   if (defined ($PARAM->{tolerance_min_band})){
     $tolerance_min_band = $PARAM->{tolerance_min_band};
   };
   if (defined ($PARAM->{tolerance_max_band})){
     $tolerance_max_band = $PARAM->{tolerance_max_band};
   };
   close (FilePar);
   
# print parameter values

   print "Parameter values:\n\n";

   print "mode                                = ",$PARAM->{mode},"\n";   
   print "shared_length_to_count_overlap      = ",$shared_length_to_count_overlap,"\n";
   print "shared_length_to_count_overlap2     = ",$shared_length_to_count_overlap2,"\n";
   print "constraint length deviation         = ",$Dev,"\n";
   print "default BAC length                  = ",$default_BAC_length,"\n";
   print "default constraint minimum size     = ",$default_constr_min_size,"\n";
   print "default constraint maximum size     = ",$default_constr_max_size,"\n";
   print "Fingerprint BAC size multiplier     = ",$multiplier,"\n";
   print "Fingerprint minimum fragment size   = ",$min_band_size,"\n";
   print "Fingerprint maximum fragment size   = ",$max_band_size,"\n";
   if (not defined($PARAM->{size_errors_info})){
     print "Tolerance for minimum fragment size = ",$tolerance_min_band,"\n";
     print "Tolerance for maximum fragment size = ",$tolerance_max_band,"\n";
   } 
   print "\n";
   
   return;
}
 
#
# main procedure
#

sub fassi_main{
    
  print "FASSI process:\n";
  
  # read fragment size errors file
  
  if (defined($PARAM->{size_errors_info})){
    print "Fragment size errors file(s): $PARAM->{size_errors_info} \n";
    fill_size_errors($PARAM->{size_errors_info});
  }
  
  # read an fpc file
  
  if (defined($PARAM->{fpc_map})){
    print "FPC map file: $PARAM->{fpc_map}\n";
    ($fpc_contigs, $fpc_clones, $parent, $kids) = readfpc($PARAM->{fpc_map});
  }
  
  # read band_sizes

  if (defined($PARAM->{sizes_files})){
    print "Band sizes file(s) : $PARAM->{sizes_files}\n";
    read_band_sizes($PARAM->{sizes_files});
  }
  

# read BAC-ends name conversion
  
  if (defined($PARAM->{BAC_end_reads})){
    print "BAC-end reads file(s) : $PARAM->{BAC_end_reads}\n";
    get_BACend_reads($PARAM->{BAC_end_reads});
  }

# read BAC-end reads and quality

  if (($PARAM->{mode} ne "LIGHT") and
      defined($PARAM->{reads_fasta}) and
      defined($PARAM->{reads_quality}))
  {
    print "Reads fasta files : $PARAM->{reads_fasta}\n";
    read_sequence($PARAM->{reads_fasta}); 

    print "Reads quality files : $PARAM->{reads_quality}\n";
    read_quality($PARAM->{reads_quality}); 
  }  



 
  ConFilesOpen(); 



# getting constraints
  if (defined($PARAM->{output_constraints})){
    print "getting constraints...\n";
    get_constraints($PARAM->{output_constraints});
  }
  
#  print reads and qualities for the new reads
  if (defined($PARAM->{output_reads}) and defined($PARAM->{output_qual})){
    print "Writing new reads and qualities...\n";
    print_reads_quality();
    print "Done\n";
  }
  
  ConFilesClose();
};



# open output files  

sub ConFilesOpen{
  open (FileReads,"> $PARAM->{output_reads}");  
  open (FileQual,"> $PARAM->{output_qual}");
  open (FileCon,"> $PARAM->{output_constraints}");
  open (FileStat,"> $PARAM->{output_stat}");
}  


#close output files

sub ConFilesClose{
  close (FileReads);  
  close (FileQual);
  close (FileCon);
  close (FileStat);
}  

# get constraints: length, buried and overlap

sub  get_constraints{
   my ($set_bands1,$set_bands2);
   my ($curr_contig,$curr_clone);
   my ($common_set,$unique_set);
   my ($clone1,$clone2,$curr_clone_length,$curr_child_length,$clone1_length,$clone2_length,$clone1_bands,$clone2_bands);
   my ($curr1_BAC_left,$curr1_BAC_right,$curr2_BAC_left,$curr2_BAC_right);
   my ($i,$curr_child);
   my ($len1,$len2);
   my ($shared_length,$unique_length,$distance);
   my $str2print;
   my $quan_constr = 0;
   my ($quan_BAC_length,$quan_buried,$quan_overlapped) = (0,0,0);
   my ($sum_BAC_length,$sum_buried,$sum_overlapped) = (0,0,0);
   my ($l,$curr_ctg_left);
   my $similarity_perc;
   my $BAC_list_for_constr = [];
   my $quan_changed;
   my $temp_unused;
 
   my ($min_constr_size,$max_constr_size);
   
   # for each BAC length and burying constraints generated 
   foreach (keys %{$fpc_clones}){
     $curr_clone = $_;
     $curr_clone_length = $default_BAC_length;
     ($min_constr_size,$max_constr_size) = ($default_constr_min_size,$default_constr_max_size);
     
     # if BES presents
     if (    
         defined($fpc_clones->{$curr_clone}{left}[0]) and
         defined($fpc_clones->{$curr_clone}{right}[0])
         ){
        if (defined($fpc_clones->{$curr_clone}{contig}) and
            $fpc_clones->{$curr_clone}{contig} != 0 and
            defined ($fpc_clones->{$curr_clone}{length})
           ){
            $curr_clone_length = $fpc_clones->{$curr_clone}{length};
            ($min_constr_size,$max_constr_size) =
             (int($curr_clone_length*(1-$Dev)),int($curr_clone_length*(1+$Dev)));

        }   
        $curr1_BAC_left = $fpc_clones->{$curr_clone}{left}[0];
        $curr1_BAC_right = $fpc_clones->{$curr_clone}{right}[0];
        $quan_BAC_length++;
        $sum_BAC_length += $curr_clone_length;
        $str2print = sprintf "%-25s %-25s %9d %9d len%1d\n",
                  $curr1_BAC_left,$curr1_BAC_right,
                  $min_constr_size,
                  $max_constr_size,
                  ++$quan_constr;
        print FileCon $str2print;
        
        # for all BACs, buried under curr_clone
        if (defined ($kids->{$curr_clone})){
          foreach(@{$kids->{$curr_clone}}){
            $curr_child = $_;
            $curr_child_length = $default_BAC_length;
            ($min_constr_size,$max_constr_size) = ($default_constr_min_size,$default_constr_max_size);
            if (defined ($fpc_clones->{$curr_child}{length})){
              $curr_child_length = $fpc_clones->{$curr_child}{length};
              ($min_constr_size,$max_constr_size) =
                (int($curr_child_length*(1-$Dev)),int($curr_child_length*(1+$Dev)));
            }    
            if ($curr_child_length >= $curr_clone_length){
              $curr_child_length = $curr_clone_length*0.9; 
            }
            if (defined ($fpc_clones->{$curr_child}{left}[0])){
               $curr2_BAC_left = $fpc_clones->{$curr_child}{left}[0];
               $str2print = sprintf "%-25s %-25s %9d %9d bur%1d\n",
                            $curr2_BAC_left,$curr1_BAC_right,
                            $min_constr_size,
                            $max_constr_size,
                            ++$quan_constr;
                print FileCon $str2print;
                $quan_buried++;               
                $sum_buried += ($curr_clone_length*(1+$Dev)+$curr_child_length*(1-$Dev))/2;
                if ($PARAM->{mode} ne "LIGHT"){
                   $reads_to_print->{$curr2_BAC_left}=""; 
                   $str2print = sprintf "%-25s %-25s %9d %9d bur%1d\n",
                                $curr1_BAC_left,"FASSI".$curr2_BAC_left,
                                0,
                                ($curr_clone_length-$curr_child_length)*(1+$Dev),
                               ++$quan_constr;
                   print FileCon $str2print;
                   $quan_buried++;
                   $sum_buried += ($curr_clone_length - $curr_child_length)*(1+$Dev)/2;                   
                }               
            }
            if (defined ($fpc_clones->{$curr_child}{right}[0])){              
              $curr2_BAC_right = $fpc_clones->{$curr_child}{right}[0];
              $str2print = sprintf "%-25s %-25s %9d %9d bur%1d\n",
                          $curr1_BAC_left,$curr2_BAC_right,
                          $curr_child_length*(1-$Dev),
                          $curr_clone_length*(1+$Dev),
                          ++$quan_constr;
              print FileCon $str2print;
              $quan_buried++;
              $sum_buried += ($curr_clone_length*(1+$Dev)+$curr_child_length*(1-$Dev))/2; 
              if ($PARAM->{mode} ne "LIGHT"){
                $reads_to_print->{$curr2_BAC_right}=""; 
#                print FileReads ">"."FASSI".$curr2_BAC_right."\n".reverseComplement($sequence->{$curr2_BAC_right})."\n";
#                print_qual($curr2_BAC_right,1);
                $str2print = sprintf "%-25s %-25s %9d %9d bur%1d\n",
                            "FASSI".$curr2_BAC_right,$curr1_BAC_right,
                            0,
                            ($curr_clone_length-$curr_child_length)*(1+$Dev),
                            ++$quan_constr;
                print FileCon $str2print;
                $quan_buried++;
                $sum_buried += ($curr_clone_length-$curr_child_length)*(1+$Dev)/2;
              }  
            }   
          }    
        }
     }
   }
   
 
   my $curr_clone1;
   my ($satisfied_overlap, $unsatisfied_overlap) = (0,0);
   my $curr_ctg_num;
   my $clone_seq_name;
   
   # overlap constraints generating
   # every contig - hash of left coordinates 
   foreach(keys %{$fpc_contigs}){
      $curr_ctg_num = $_;
      $curr_contig = $fpc_contigs->{$_};
      $i = 0;
      $BAC_list_for_constr = [];
      # for every left border
      foreach( sort {$a <=> $b} keys %{$curr_contig}){
        $curr_ctg_left = $curr_contig->{$_};
        #for every BAC. As a result we have list of BACs, ordered by left end
        foreach( sort {$a <=> $b} keys %{$curr_ctg_left}){
          $curr_clone = $curr_ctg_left->{$_};
          if (defined ($fpc_clones->{$curr_clone}{left}[0]) and
              defined ($fpc_clones->{$curr_clone}{right}[0]) and
              defined ($fpc_clones->{$curr_clone}{length})){
            my $curr_clone1 = {(name => $curr_clone,
                             bands => $band_sizes->{$curr_clone},
                             length => $fpc_clones->{$curr_clone}{length},
                             left => $fpc_clones->{$curr_clone}{left}[0],
                             right => $fpc_clones->{$curr_clone}{right}[0],
                             )};


            
            
            $similarity_perc = $shared_length_to_count_overlap;
            for (my $k = @$BAC_list_for_constr-1; $k>=0;$k--){
              $_ = $BAC_list_for_constr->[$k];
              ($unique_set,$temp_unused,$common_set) = set_bands_comparison($_->{bands},$curr_clone1->{bands},0);
              $shared_length = get_clone_length($common_set);
              if ($shared_length >= $_->{length}*0.95){
                $shared_length = $_->{length}*0.95;
              }
              if ($shared_length >= $curr_clone1->{length}*0.95){
                $shared_length = $curr_clone1->{length}*0.95;
              }  
              $unique_length = $_->{length} - $shared_length;
              if ($shared_length > $unique_length*$similarity_perc){
                $satisfied_overlap++;
                $str2print = sprintf "%-25s %-25s %9d %9d over%1d\n",
                          $_->{left},$curr_clone1->{right},
                          int(($_->{length}+$curr_clone1->{length}-$shared_length)*(1-$Dev)),
                          int(($_->{length}+$curr_clone1->{length}-$shared_length)*(1+$Dev)),
                          ++$quan_constr;
                print FileCon $str2print;
                $str2print = sprintf "%-25s %-25s %9d %9d over%1d\n",
                      $curr_clone1->{left},$_->{right},
                      int($shared_length*(1-$Dev)),
                      int($shared_length*(1+$Dev)),
                      ++$quan_constr;
                print FileCon $str2print;
                $quan_overlapped+=2;
                $sum_overlapped += ($_->{length}+$curr_clone1->{length}-$shared_length);
                $sum_overlapped += $shared_length;
                if ($PARAM->{mode} ne "LIGHT"){
                  if ($k==@$BAC_list_for_constr-1){
                    $reads_to_print->{$curr_clone1->{left}}="";
                    $reads_to_print->{$curr_clone1->{right}}=""; 
                  };      
                  $str2print = sprintf "%-25s %-25s %9d %9d over%1d\n",
                               $_->{left},"FASSI".$curr_clone1->{left},
                               int(($_->{length}-$shared_length)*(1-$Dev)),
                               int(($_->{length}-$shared_length)*(1+$Dev)),
                               ++$quan_constr;
                  print FileCon $str2print;

                  $str2print = sprintf "%-25s %-25s %9d %9d over%1d\n",
                              "FASSI".$_->{right},$curr_clone1->{right},
                              int(($curr_clone1->{length}-$shared_length)*(1-$Dev)),
                              int(($curr_clone1->{length}-$shared_length)*(1+$Dev)),
                              ++$quan_constr;
                  print FileCon $str2print;
                  $quan_overlapped+=2;
                  $sum_overlapped += ($_->{length}-$shared_length);
                  $sum_overlapped += ($curr_clone1->{length}-$shared_length);
                };  
                if ($PARAM->{mode} ne "STRONG" ){
                  shift @$BAC_list_for_constr;  
                }    
              } else {
                $unsatisfied_overlap++;
                for (my $j=0;$j<$k+1;$j++){
                  shift @$BAC_list_for_constr;  
                };    
                last;
              }
              $similarity_perc = $shared_length_to_count_overlap2;
            };
            push (@{$BAC_list_for_constr},$curr_clone1);
          };   
        };
      };      
   };
   if ($quan_BAC_length == 0){
     print FileStat "BACs       : 0 , average size: 0\n";
     print "BACs       : 0 , average size: 0\n";
   } else {
     print FileStat "BACs       : ",$quan_BAC_length," , average size: ",int($sum_BAC_length/$quan_BAC_length),"\n";
     print "BACs       : ",$quan_BAC_length," , average size: ",int($sum_BAC_length/$quan_BAC_length),"\n";
   } 
   if ($quan_buried == 0){
     print FileStat "Burying    : 0 , average size: 0\n";    
     print "Burying    : 0 , average size: 0\n";    
   } else { 
     print FileStat "Burying    : ",$quan_buried," , average size: ",int($sum_buried/$quan_buried),"\n";
     print "Burying    : ",$quan_buried," , average size: ",int($sum_buried/$quan_buried),"\n";
   }
   if ($quan_overlapped == 0){
     print FileStat "Overlapped : 0 , average size: 0\n";
     print "Overlapped : 0 , average size: 0\n";
   } else {
     print FileStat "Overlapped : ",$quan_overlapped," , average size: ",int($sum_overlapped/$quan_overlapped),"\n";
     print "Overlapped : ",$quan_overlapped," , average size: ",int($sum_overlapped/$quan_overlapped),"\n";
   }; 
   print "Satisfied overlaps: ",$satisfied_overlap, "    Unsatisfied overlaps: ",$unsatisfied_overlap,"\n";
   
   return;
};




#
# 2 band sets comparison. Returns unique bands for
# every set and coomon bands with recalculated edges   
#

sub set_bands_comparison {
   my ($unique_bands1,$unique_bands2,$common_bands)=([],[],[]);
   my ($bands_set1,$bands_set2,$recalc_edges_mode) = @_;
   my ($band1, $band2);
   my ($left, $right);
   my ($middle_band1);
   my ($pos1, $pos2) = (0,0);
   my ($len1, $len2) = (scalar(@$bands_set1),scalar(@$bands_set2));
   $band2 = $bands_set2->[$pos2];
   while ($pos1 < $len1){
     $band1 = $bands_set1->[$pos1];
     $middle_band1 = ($band1->[0]+$band1->[1])/2;
     while ($pos2 < $len2 and $middle_band1 > $band2->[1]){
       push (@$unique_bands2,$band2);
       $pos2++;
       if ($pos2 < $len2){
         $band2 = $bands_set2->[$pos2];
       };  
     }
     if ($pos2 < $len2 and $middle_band1 > $band2->[0]){
       if ($recalc_edges_mode == 1){
         $left = $band2->[0];
         $right = $band2->[1];
       } elsif ($recalc_edges_mode == 2){
         $left = ($band1->[0]+$band2->[0])/2;
         $right = ($band1->[1]+$band2->[1])/2;
       } else {   
         $left = $band1->[0];
         $right = $band1->[1];
       }    
       push(@$common_bands,[($left,$right)]);
       $pos2++;
       if ($pos2 < $len2){
         $band2 = $bands_set2->[$pos2];
       };  
     } else {
       push (@$unique_bands1,$band1); 
     }   
     $pos1++;
   }
   while ($pos2 < $len2){
     push (@$unique_bands2,$bands_set2->[$pos2++]);
   }
   return ($unique_bands1,$unique_bands2,$common_bands);
}


# getting BAC length based on bands sizes

sub get_clone_length {
   my $bands = $_[0];
   my $length=0;
   foreach (@$bands) {
     $length += (($_->[0]+$_->[1])/2) %  $max_band_size;
   }
   return $length * $multiplier;
}


#
# read all the band sizes
# Parameter:
# $dirname - path to the folder with band sizes files
#


sub read_band_sizes {
   my @sizes_folder = @_;
   my $band_sizes_filename;
   my ($curr_BAC,$BAC_length);
   my $size_err_item = 0;
   my ($a_tol,$b_tol,$tolerance) = (0,0,0);
   if (not defined($PARAM->{size_errors_info})){
     $a_tol = ($tolerance_max_band - $tolerance_min_band)/
                ($max_band_size - $min_band_size);
     $b_tol = $tolerance_max_band - $a_tol*$max_band_size;
   } 
   while ($band_sizes_filename = <@sizes_folder>){
     open (band_sizes, $band_sizes_filename) or die "can't open the file $band_sizes_filename\n";
     $curr_BAC = "";
     while (<band_sizes>) {
       if (/^\s*(\S+)\s+(\d+)/) {
         $curr_BAC = $1;
         $BAC_length = 0;
       } else {
          if (/^(-?\d+\.?\d*)/){
            if ($1 == -1){
              $fpc_clones->{$curr_BAC}{length} = $BAC_length*$multiplier;   
            } else {
              $BAC_length += $_ % $max_band_size;
              if (defined($PARAM->{size_errors_info})){
                if ($1 < $size_errors_min){
                  $size_err_item = $size_errors_min }
                elsif ($1 > $size_errors_max){
                    $size_err_item = $size_errors_max }
                else { $size_err_item = $1 };
                push (@{$band_sizes->{$curr_BAC}},[$1*(1-$size_errors->[$size_err_item]),$1*(1+$size_errors->[$size_err_item])]);
              } else {
                $tolerance = $a_tol*($1 % $max_band_size) + $b_tol;
                push (@{$band_sizes->{$curr_BAC}},[$1-$tolerance,$1+$tolerance]);
              }  
            }  
          }
       }    
     }  
     close (band_sizes);
   }
   return;
};




#
# read FPC file.
# Parameters:
# $fpcfile - full name of file with fpc map
#
sub readfpc {
   my $fpcfile = $_[0];
   my $header = "";
   my $footer = "";
   open (FPC, $fpcfile) or die "can't open the file $fpcfile\n";
   my (%contigs_to_return,%clones_to_return,%parents_to_return,%kids_to_return);
   my $ncontigs = 0;
   my $nBACs = 0;
# read the header
   while (<FPC>) {
      unless (/\/\//) {
         last;
      } else {
         $header .= $_;
	 if (/Contigs (\d+)/) { $ncontigs = $1;}
      }
   }
   $header .= "\n";
# read the clones
   my $currentclone = "";
   my $curr_contig="";
   my ($curr_leftposition, $curr_rightposition);
   my $is_buried=0;
   my $is_canceled=0;

   while (<FPC>) {
      if (/Markerdata/) {
         $footer .= $_;
         last;
      } else {
         if ( /(Clone : |BAC : |YAC : |Cosmid :|Fosmid :|PAC :)\"(\S+)\"/ ) {
            ($curr_contig,$curr_leftposition,$curr_rightposition) = (0,0,0);
	    $currentclone = $2;
            $is_buried=0;
            $is_canceled = 0;
	    if ($currentclone =~ /!/) {
	       $is_canceled = 1;
	    }
	 } elsif (/^Map\s+\"ctg(\d+)\"\s+Ends Left\s+(-?\d+)/) {
            $curr_contig = $1;
            $curr_leftposition = $2;
	 } elsif (/^Map\s+\"ctg(\d+)\"\s+Ends Right\s+(-?\d+)/) {
            $curr_rightposition = $2;
	 } elsif ( /match_to_cosmid\s+\"(\S+)"/ ) {
            $is_buried=1;
            $parents_to_return{$currentclone} = $1;
            push (@{$kids_to_return{$1}},$currentclone);
	 } elsif ( /^Creation_date/ ) {
            if (!$is_canceled){
              $clones_to_return{$currentclone}{contig}=$curr_contig;
              $clones_to_return{$currentclone}{ctg_left}=$curr_leftposition;
              $clones_to_return{$currentclone}{ctg_right}=$curr_rightposition;
              $nBACs++;
              if (!$is_buried and $curr_contig != 0) {
                $contigs_to_return{$curr_contig}{$curr_leftposition}{$curr_rightposition} = $currentclone;
              }  
            }
	 }
         else {

	 }
      }
   }
# read the marker data
#   while (<FPC>) {
#if (/Contigdata/) {last;}    # don't want the Contigdata in the output
#     $footer .= $_;
#   }
   close (FPC);
   print "BACs: ",$nBACs," Contigs: ",$ncontigs,"\n";
   print "Same position in contig: ",$num_same_position,"\n";
   return \(%contigs_to_return,%clones_to_return,%parents_to_return,%kids_to_return);
}


# filling array of errors in band sizes

sub fill_size_errors{
   my $file_size_errors = shift; 
   open (FileSizeErr, $file_size_errors) or die "can't open the file $file_size_errors\n";
   while (<FileSizeErr>) {
      if ( /^(\d+)\s+(\d+.\d+)/ ) {
        $size_errors->[$1] = $2;
      }
   }
   close (FileSizeErr);
   return;
}




# make reverse-compliment reads

sub reverseComplement{

    #Reverse complements a DNA sequence
    #only accepts sequence as string
    #returns the reverse complemented sequence in a string

    my $seq=shift;
    my @temp_array=split(//,$seq);
    my @temp_array1=reverse(@temp_array);
    my $rev_seq="";

    #reverse complements array
    my %comprev = ("G","C","C","G","A","T","T","A","g","c","c","g","a","t","t","a");

    for (my $i=0; $i<=$#temp_array1; $i++){
        if (defined $comprev{$temp_array1[$i]}){
            $rev_seq.=$comprev{$temp_array1[$i]};
        }
        else{
            $rev_seq.="N";
        }
    }
    return $rev_seq;
}

# read fasta quality files

sub read_quality{
  my ($read_name,$new_read_name);
  my $qualfilename;
  my $cur_qual_str;
  while ($qualfilename = <@_>){
    open (Qual, $qualfilename) or die "can't open the file $qualfilename\n";
    $cur_qual_str = "";
    while (<Qual>) {
      if (/^>(\S+)/){
        $new_read_name = $1;
        if (defined($read_name) and $read_name ne "" and 
            defined($sequence->{$read_name})){
            $quality->{$read_name} = $cur_qual_str;
        }
        $cur_qual_str = "";
        $read_name = $new_read_name;
      }  
      else {
        $cur_qual_str .= $_;
      }
    }
    if (defined($read_name) and not($read_name eq "")){
       if (defined($sequence->{$read_name})){
          $quality->{$read_name} = $cur_qual_str;        
       };  
    }
    close (Qual);    
  }       
  return;
}

# print reads and qualities to files in fasta format

sub print_reads_quality {
  foreach (keys %$reads_to_print){
    print FileReads ">"."FASSI".$_."\n".reverseComplement($sequence->{$_})."\n";
    print_qual($_,1);
  }    
}    

# print quality for particular BAC either in straight or reversed order

sub print_qual {
  my ($curr_read,$reverse) = @_;
  my $i;
  my $curr_qual_str;
  my $str_to_print;
  if (defined($quality->{$curr_read})){
    if ($reverse == 0){
      print FileQual ">".$curr_read."\n";  
      print FileQual $quality->{$curr_read};  
    } else {
      print FileQual ">"."FASSI".$curr_read."\n";
      $curr_qual_str = reverse $quality->{$curr_read};
      while ($curr_qual_str =~ /(\S+)/){
        $str_to_print = reverse $1;
        $curr_qual_str = $';
        print FileQual $str_to_print," ";
      }
      print FileQual "\n";
    }    
 };  
  
}

#
# Read BAC-end read names from file with format:
# <Fingerprint BAC-name> <"left" BAC-end readname> <"right" BAC-end readname>
#
sub get_BACend_reads{
  my $fileBACends = shift;
  my $BAC_name;
  my ($i,$j) = (0,0);
  open (FileBACendReads,$fileBACends) or die "can't open the file $fileBACends\n";
  while (<FileBACendReads>) {
    if (/^(\S+)\s+(\S+)\s+(\S+)/){
      $i++;  
      if (not defined($fpc_clones->{$1})){ $j++};  
      push(@{$fpc_clones->{$1}{left}},$2);
      push(@{$fpc_clones->{$1}{right}},$3);
      $sequence->{$2}="";
      $sequence->{$3}="";
    }  
  }
  close (FileBACendReads);
  print "Load ",$i," BACs; ",$j," of them - new.\n";
  return;
};    
    

# read sequence from fasta-file

sub read_sequence{
  my $read_name="";
  my $bac_name;
  my $bacfilename;
  my $temp_bac_list;
  my $cur_seq_str;
  while ($bacfilename = <@_>){
    open (Seq, $bacfilename) or die "can't open the file $bacfilename\n";
    $cur_seq_str = "";
    while (<Seq>) {
      if (/^>(\S+)/){
        if ($read_name ne "" and defined($sequence->{$read_name})){
          $sequence->{$read_name} = $cur_seq_str;
        }
        $cur_seq_str = "";
        $read_name = $1;
      }  
      else {
        chomp;
        $cur_seq_str .= $_;
      }
    }
    if ($read_name ne "" and defined($sequence->{$read_name})){
      $sequence->{$read_name} = $cur_seq_str;
    }
    close (Seq);    
  }       
  return;
}

#
# help message
#
format STDOUT =
Usage: Fassi <profile-filename>
profile examples - fly.profile, chimp.profile
.

